-----------------------------------------------------------------------------
--Gif file decoding routines Mike Carroll 12/4/97

--A Gif file starts with a Gifheader structure, possibly followed by a 
--global colour table. This is followed by a single byte identifier char
--of a ',' for a image following or a '!' for an extension block.
--Any extension blocks found are skipped by this decoder.
--An imageblock starts with an image header structure possibly followed by
--a local colour table and any number of data blocks each preceeded by a
--single byte defining the size of the block.
--More than one image can be stored in a gif file, the end of the gif is
--terminated by a ';' char.
--This decoder will only return the first image in the file.

--Function unPackImage which does the LZW decompression of the actual
--image data is based on a 'C' routine which appears in the book 
--'Bit-Mapped Graphics' by Steve Rimmer. The author of this book does not
--provide any comments with his routine mearly stating that its based on
--code found on a bulletin board, and the author's name was not included in 
--the file.

-----------------------------------------------------------------------------
-- changed read_gif output to standard {pallete,image}  - Paul Plummer
-- changed palette to 8 bit color to match read_bitmap
-- added function info_gif(sequence filename)  
-- returns sequence {number of colors,width,height} or atom errorcode

include get.e
include file.e
include image.e

without type_check

global constant NO_CODE = -1, TRUE = 1, EOF = -1,
                                ERROR_EOF = 1,ERROR_BAD_CODE = 6,
                                ERROR_BAD_HEADER = 3,ERROR_BAD_STARTCODE = 4,
                                ERROR_BAD_FIRST_CODE = 5,ERROR_BAD_FILE = 2, 
                                ERROR_NO_IMAGE = 7

global sequence GifErrorString

GifErrorString = {
                "Unexpected end of file\n",
                "Error opening file\n",
                "Bad gif header\n",
                "Bad symbol size\n",
                "Bad first code\n",
                "Bad code\n",
                "This file doesn't contain an image\n"}

--GIFHEADER structure
constant SIG = 1,       --File signature has to be 'GIF87a' or 'GIF89a'
                 SCREENWIDTH = 2,--Logical Screen Width
                 SCREENDEPTH = 3,--Logical Screen Height
                 HFLAGS = 4,     --Global Color Table Flag       1 Bit           MSB
                                                 --Colour Resolution             3 Bits
                                                 --Sort Flag                     1 Bit
                                                 --Size of Global Colour Table   3 Bits          LSB
      --BACKGROUND = 5, --Index into the Global Colour Table for the 
                                                 --Background Colour
                 ASPECT = 6      --Pixel Aspect Ratio

--IMAGEBLOCK structure
constant          --LEFT = 1,   --Image Left Position
                          --TOP = 2,    --Image Top Position
                          --WIDTH = 3,  --Image Width
                          --DEPTH = 4,  --Image Height
                      IFLAGS = 5  --Local Color Table Flag        1 Bit           MSB
                                                  --Interlace Flag                1 Bit
                                                  --Sort Flag                     1 Bit
                                                  --Reserved                      2 Bits
                                                  --Size of Local Color Table     3 Bits          LSB

--      ImageInfo structure returned by GifLoad
constant         
        LEFT = 1,       --Image Left Position
        TOP = 2,        --Image Top Position                            
        WIDTH = 3,      --Image Width
        DEPTH = 4,      --Image Height
        BACKGROUND = 5, --Background colour
        PALETTE = 6,    --Image colour palette
        BITS = 7,       --Size of Color Table
        IMAGE = 8       --Image data

---------------------------------------------------------------------------
function ShiftRight(integer a,integer bits)
    return floor(a / power(2,bits))
end function

---------------------------------------------------------------------------
function ShiftLeft(integer a, integer bits)
    return a*power(2,bits)
end function

---------------------------------------------------------------------------
--read colour palette, vga palette values are 6 bit numbers
--while gif allows for 8 bit so shift right to get correct colours

function ReadPalette(integer fn, integer bytes)
    sequence block

    block = repeat({0,0,0},(bytes / 3))
    for a = 1 to (bytes / 3) do
        for b = 1 to 3 do
--            block[a][b] = ShiftRight(getc(fn),2)
            block[a][b] = getc(fn)
        end for
    end for
    return block
end function

---------------------------------------------------------------------------
--read a block of bytes into memory

function BlockMemRead(integer fn, atom buffer, integer bytes)
    atom a
    for cnt = 0 to bytes-1 do
                a = getc(fn)
                if a = EOF then
                    return EOF
                else
            poke(buffer+cnt,a)
                end if
    end for
    return TRUE
end function 

---------------------------------------------------------------------------
--read a unsigned 16 bit value from file, low byte first

function ReadWord( integer fn )
    atom a, b

    a = getc(fn)
    b = getc(fn)
    return (b * 256 + a)
end function

---------------------------------------------------------------------------
--read the GIF file header structure into a sequence

function GetGifHeader( integer fn )
    sequence Header

    Header = {{0,0,0,0,0,0},0,0,0,0,0}
    for cnt = 1 to 6 do
        Header[SIG][cnt] = getc(fn)
    end for

     --test for valid file, first three bytes must be 'GIF'
    if compare(Header[SIG][1..3],"GIF") != 0 then
                return ERROR_BAD_HEADER --Error reading GIF header
    end if
    Header[SCREENWIDTH] =ReadWord(fn)       
    Header[SCREENDEPTH] =ReadWord(fn)       
    Header[HFLAGS] = getc(fn)               
    Header[BACKGROUND] = getc(fn)           
    Header[ASPECT] = getc(fn)
    return Header
end function

global function info_gif(sequence fname)
object Header
integer fn,colors
fn=open(fname,"rb")
    if fn = -1  then  --Error opening file
                return 2
    end if
Header=GetGifHeader(fn)
close(fn) 
if atom(Header) then
return Header
else    
colors= power(2,and_bits(Header[HFLAGS],#7)+1)
return {colors,Header[SCREENWIDTH],Header[SCREENDEPTH]}
end if
end function
---------------------------------------------------------------------------
--gif file can contain more than one image,
-- each image is preceeded by a header structure

function GetImageHeader( integer fn )
    sequence Header

    Header = {0,0,0,0,0}
    Header[LEFT] = ReadWord( fn)
    Header[TOP] = ReadWord( fn)
    Header[WIDTH] = ReadWord(fn)
    Header[DEPTH] = ReadWord(fn)
    Header[IFLAGS] = getc(fn)
    return Header
end function

---------------------------------------------------------------------------
--unpack an LZW compressed image 
--returns a sequence containing screen display lines of the image

function unPackImage(integer fn, integer startCodeSize, 
                        integer width,integer depth, integer flags )
    sequence wordMaskTable,
            buffer,         --image buffer
            lineBuffer,     --place to store the current line of the image
            incTable,       --interlace increments
            startTable,     --interlace starts 
            firstCodeStack, --stack for first codes 
            lastCodeStack,  --stack for previous codes
            codeStack       --stack for links

    integer bitsLeft,   --number of bits left in read buffer
            clearCode,      --clear code
            codeSize,       --current code size in bits
            codeSize2,      --next code size in bits
            nextCode,       --next available table entry
            thisCode,       --code being expanded
            oldToken,       --last symbol decoded
            currentCode,    --code just read
            oldCode,        --code read before this one
            blockSize,      --bytes in next block
            line,           --next line to write
            byte,           --next byte to write
            pass,           --pass number for interlaced pictures
            u               --stack pointer into firstCodeStack

    atom b,             --read buffer 
            p,              --current byte in read buffer
            q               --last byte in read buffer + 1

    wordMaskTable = {       
                    #0000, #0001, #0003, #0007,
                    #000F, #001F, #003F, #007F,
                    #00FF, #01FF, #03FF, #07FF,
                    #0FFF, #1FFF, #3FFF, #7FFF 
                    }

    incTable = { 8,8,4,2,0 }                --interlace increments
    startTable = { 0,4,2,1,0 }              --interlace starts

    lineBuffer  = repeat(0,width)
    firstCodeStack = repeat(0,4096)
    lastCodeStack = repeat(0,4096)
    codeStack = repeat(0,4096)
    buffer = repeat({},depth)

    byte = 1 line = 0 pass = 1
    bitsLeft = 8
    if startCodeSize < 2 or startCodeSize > 8 then
                return ERROR_BAD_STARTCODE --Bad symbol size
    end if

    b = allocate(256)               --buffer for block reads
    p = b
    q = b
    clearCode = power(2,startCodeSize)
    nextCode = clearCode + 2
    codeSize = startCodeSize + 1
    codeSize2 = power(2,codeSize)
    oldCode = NO_CODE
    oldToken = NO_CODE

    --gif image data is stored in blocks preceeded by a byte containing the
    --block size, each code is between 3 and 12 bits in length packed into
    --8 bit bytes

    while TRUE do
                if bitsLeft = 8 then
                    p = p + 1            --point to next byte in buffer
                    if p >= q then       --if were at the end then get next block
                                blockSize = getc(fn)    --bytes in this block
                                if BlockMemRead(fn,b,blockSize)= EOF then
                                    free(b)
                                    return ERROR_EOF --Unexpected end of file
                                end if
                                p = b                   --reset pointer to start of buffer
                                q = b + blockSize --end of buffer marker
                    end if
                    bitsLeft = 0 
                end if
                thisCode = peek(p)              --get low data byte
                currentCode = codeSize + bitsLeft

                --if currentCode < 8 then some bits of the next code are in this byte, 
                --move these bits into the LSB ready for next code read
                if currentCode <= 8 then
                    poke(p,ShiftRight(peek(p),codeSize))
                    bitsLeft = currentCode
                else
                           --if currentCode > 8 then then some bits of this code 
                    --are in the next byte so get it
                    p = p + 1
                    if (p >= q) then
                                blockSize = getc(fn)
                                if BlockMemRead(fn,b,blockSize)= EOF then
                                    free(b)
                                    return ERROR_EOF --Unexpected end of file
                                end if
                                p = b
                                q = b + blockSize
                    end if
                    --bits in next byte are high order bits of this code
                    --so move to MSB of code and OR in
                    thisCode = or_bits(thisCode,ShiftLeft(peek(p),(8 - bitsLeft)))

                        --if currentCode < 16 then some bits of the next code are in this byte, 
                        --move these bits into the LSB ready for next code read
                    if currentCode <= 16 then
                                bitsLeft = currentCode - 8
                                poke(p,ShiftRight(peek(p),bitsLeft))
                    else
                                --if currentCode > 16 then then some bits of this code 
                                --are in the next byte so get it
                                p = p + 1
                                if (p >= q) then
                                    blockSize = getc(fn)
                                        if BlockMemRead(fn,b,blockSize)= EOF then
                                                free(b)
                                                return ERROR_EOF --Unexpected end of file
                                    end if
                                    p = b
                                    q = b + blockSize
                                end if

                                --bits in next byte are high order bits of this code
                                --so move to MSB of code and OR in
                                thisCode = or_bits(thisCode,ShiftLeft(peek(p),(16 - bitsLeft)))

                                --if there is any bits of next code left then
                                --move these bits into the LSB ready for next code read
                                bitsLeft = currentCode - 16
                                poke(p,ShiftRight(peek(p),bitsLeft))
                           end if
                end if

                --blank out the bits of the next code from this code
                thisCode = and_bits(thisCode, wordMaskTable[codeSize+1])
                currentCode = thisCode --save the new code

                --blocksize = 0 is really an error but some gifs use
                --this as an end of image marker
                if thisCode = (clearCode+1)or blockSize = 0 then
                    --if we get here then we've read the image
                    --and can return, jump out of loop and tidy up
                    exit
                end if

                if thisCode > nextCode then --Bad code
                    --the file must be corrupt 
                    free(b)
                        return ERROR_BAD_CODE
                end if

                if thisCode = clearCode then
                    --instruction to trash tables and start again
                    nextCode = clearCode + 2
                    codeSize = startCodeSize + 1
                    codeSize2 =  power(2,codeSize)
                    oldCode = NO_CODE
                           oldToken = NO_CODE
                else
                    --must be valid code so decode it
                    u = 1
                    if thisCode = nextCode then
                                if oldCode = NO_CODE then  --Bad first code
                                    --first code read must be less than 256 & in table
                                    free(b)
                                        return ERROR_BAD_FIRST_CODE
                                end if

                                --codes are in reverse order so store in LIFO stack
                                firstCodeStack[u] = oldToken
                                u = u + 1
                                thisCode = oldCode
                    end if

                    while thisCode >= clearCode do
                                firstCodeStack[u] = lastCodeStack[thisCode]
                                u = u + 1
                                thisCode = codeStack[thisCode]
                    end while

                    oldToken = thisCode
                    while TRUE do
                                lineBuffer[byte] = thisCode
                                byte = byte + 1
                                if byte > width then
                                           --full image line so add it into screen image
                                    buffer[line+1] = lineBuffer
                                    byte = 1                --reset line byte count

                                    if and_bits(flags, #40) then
                                                --interlace flag set so calc next line posn
                                                line = line + incTable[pass]
                                                if line >= depth then
                                                    pass = pass + 1
                                                    line = startTable[pass]
                                                end if
                                    else
                                                --not interlaced
                                                line = line + 1
                                    end if
                                end if
                                if u = 1 then
                                    --no more bytes on stack
                                    exit
                                end if
                                --pop next byte from LIFO stack
                                u = u - 1
                                thisCode = firstCodeStack[u]
                    end while

                    if nextCode < 4096 and oldCode != NO_CODE then
                                codeStack[nextCode] = oldCode
                                lastCodeStack[nextCode] = oldToken
                                nextCode = nextCode + 1
                                if nextCode >= codeSize2 and codeSize < 12 then
                                    --code exceeds current code length so bump code length
                                    codeSize = codeSize + 1
                                    codeSize2 = power(2,codeSize)
                                end if
                    end if
                    oldCode = currentCode --retrieve the code
                end if
           end while

    --completed reading the image so return it
    free(b)
    return buffer
end function

---------------------------------------------------------------------------
--skip the extension blocks as we are only after the image

procedure doExtension(integer fn)
    integer n, c

    --puts(1,"Skipping an extension\n")
    n = getc(fn)                    --throwaway extension function code
    n = getc(fn)                    --get length of block
    while n > 0 and n != EOF do
                for i = 1 to n do           --throwaway the block
                    c = getc(fn)
                end for
                n = getc(fn)                --get length of next block
    end while
end procedure

---------------------------------------------------------------------------
--unpack the GIF file
--returns ImageInfo sequence containing image and image data

global function read_gif( sequence fileName )
    integer fn, c, b
        object GifHeader, ImageInfo
    sequence ImageBlock

    ImageInfo = {0,0,0,0,0,{},0,{}}

    fn = open(fileName , "rb")
    if fn = -1  then  --Error opening file
                return ERROR_BAD_FILE
    end if

    --file starts with the Logical Screen Descriptor structure
    GifHeader = GetGifHeader( fn )
        if atom(GifHeader) then -- bad gif header
                return GifHeader
        end if

    --Size of Global Color Table    
    ImageInfo[BITS]= and_bits(GifHeader[HFLAGS],#7)+1
    ImageInfo[BACKGROUND] = GifHeader[BACKGROUND]

    --get Global colour palette if there is one
    if and_bits(GifHeader[HFLAGS],#80) then         --is flags bit 8 set
                c = 3 * power(2,ImageInfo[BITS])        --size of global colour map
                ImageInfo[PALETTE]= ReadPalette(fn, c)
    end if
    c = getc(fn)

    while atom(c = #2C or c = #21 or c = 0) do
                if c = #2C then         --image seperator so unpack the image
                    ImageBlock = GetImageHeader(fn) --get the Image Descriptor
                    --if there is a local Color Table then overwrite the global table
                    if and_bits(ImageBlock[IFLAGS],#80) then
                                ImageInfo[BITS]= and_bits(ImageBlock[IFLAGS],#7)+1
                                b = 3 * power(2,ImageInfo[BITS])  --size of local colour map
                                ImageInfo[PALETTE]= ReadPalette(fn,b)
                    end if
                    c = getc(fn)     --get the LZW Minimum Code Size
        
                    ImageInfo[IMAGE] = unPackImage(fn,c,ImageBlock[WIDTH],
                                                ImageBlock[DEPTH],ImageBlock[IFLAGS])
                    close( fn )
                        if atom(ImageInfo[IMAGE]) then --error reading image
                                return ImageInfo[IMAGE]
                        end if
                
                    ImageInfo[WIDTH] = ImageBlock[WIDTH]    --pixel width of image
                    ImageInfo[DEPTH] = ImageBlock[DEPTH]    --pixel depth of image  
                    return {ImageInfo[6],ImageInfo[8]}                        --return imagedata

                elsif c = #21 then                  --Extension Introducer 
                    doExtension(fn)                 --throw the extension away
                end if
            c = getc(fn)
    end while

    return ERROR_NO_IMAGE --no image if we get here 
end function



